package muesli

import "fmt"
import "log"

type LogTracer struct {
}

func (t LogTracer) Trace(vm VM, ast Ast, val ... Value) bool {
	log.Printf("Trace: %s -> %v", ast.String(), val)
	return false
}


type FmtTracer struct {
}

func (t FmtTracer) Trace(vm VM, ast Ast, args ... Value) bool {
	arg  := args[0]
	rest := args[1:len(args)]
	fmt.Printf("Trace: %s: %v -> ", ast.String(), arg)
	for _, v := range rest {
		fmt.Printf("%v, ", v)
	}
	fmt.Printf("<-\n ")
	return false
}

func printf(vm *VM, args ...Value) []Value {
	var format string
	rest, err := ParseArgs(args, &format)
	if err != nil {
		return Fail(err)
	}
	fmt.Printf(format, rest...)
	return None()
}

func println(vm *VM, args ...Value) []Value {
	var msg string
	_, err := ParseArgs(args, &msg)
	if err != nil {
		return Fail(err)
	} else {
		fmt.Println(msg)
	}
	return None()
}

func p(vm *VM, args ...Value) []Value {	
	for _, arg := range args {
		fmt.Printf("%v\n", arg)
	}	
	return None()
}

func trace(vm *VM, args ...Value) []Value {
	var b bool = true
	fmt.Printf("command trace: %v\n", args)
	_, err := ParseArgs(args, &b)
	if err != nil {
		fmt.Printf("Error: %s\n", err.Error())
		return Fail(err)
	}
	fmt.Printf("command trace: bool: %v\n", b)
	if b {
		vm.Tracer = FmtTracer{}
	} else {
		vm.Tracer = nil
	}
	return Ok(BoolValue(b))
}

func sumi(vm *VM, args ...Value) []Value {
	slice, err := ParseArgsToIntSlice(args)
	if err != nil {
		fmt.Printf("Error: %s\n", err.Error())
		return Fail(err)
	}
	res := int(0)
	for _, val := range slice {
		res += val
	} 
	return IntOk(res)
}

func sumf(vm *VM, args ...Value) []Value {
	slice, err := ParseArgsToFloat64Slice(args)
	if err != nil {
		fmt.Printf("Error: %s\n", err.Error())
		return Fail(err)
	}
	res := float64(0)
	for _, val := range slice {
		res += val
	} 
	return FloatOk(res)
}

func addi(vm *VM, args ...Value) []Value {
	var v1, v2 int
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return IntOk(v1 + v2)
}


func addf(vm *VM, args ...Value) []Value {
	var v1, v2 float64
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return FloatOk(v1 + v2)
}

func subi(vm *VM, args ...Value) []Value {
	var v1, v2 int
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return IntOk(v1 - v2)
}


func subf(vm *VM, args ...Value) []Value {
	var v1, v2 float64
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return FloatOk(v1 - v2)
}

func muli(vm *VM, args ...Value) []Value {
	var v1, v2 int
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return IntOk(v1 * v2)
}


func mulf(vm *VM, args ...Value) []Value {
	var v1, v2 float64
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return FloatOk(v1 * v2)
}


func divi(vm *VM, args ...Value) []Value {
	var v1, v2 int
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return IntOk(v1 / v2)
}


func divf(vm *VM, args ...Value) []Value {
	var v1, v2 float64
	_, err := ParseArgs(args, &v1, &v2)
	if err != nil {
		return Fail(err)
	}	
	return FloatOk(v1 / v2)
}


func val(vm *VM, args ...Value) []Value {
	if len(args) < 1 {
		return []Value{NewErrorValuef("val requres at least one argument.")}
	}
	return args
}


func (vm *VM) RegisterBuiltins() {
	vm.RegisterBuiltin("addi", addi)
	vm.RegisterBuiltin("addf", addf)
	vm.RegisterBuiltin("sumi", sumi)
	vm.RegisterBuiltin("sumf", sumf)
	vm.RegisterBuiltin("subi", subi)
	vm.RegisterBuiltin("subf", subf)
	vm.RegisterBuiltin("divi", divi)
	vm.RegisterBuiltin("divf", divf)
	vm.RegisterBuiltin("muli", muli)
	vm.RegisterBuiltin("mulf", mulf)
	err := vm.AddOverload("mul", "mulf", FloatTypeValue, FloatTypeValue)
	if err != nil {
		fmt.Printf("Errror registering overload: %s", err)
	}
	err = vm.AddOverload("mul", "muli", IntTypeValue, IntTypeValue)
	if err != nil {
		fmt.Printf("Errror registering overload: %s", err)
	}
	err = vm.AddOverload("mul", "mulf", FloatTypeValue, IntTypeValue)
	if err != nil {
		fmt.Printf("Errror registering overload: %s", err)
	}
	err = vm.AddOverload("mul", "mulf", IntTypeValue, FloatTypeValue)
	if err != nil {
		fmt.Printf("Errror registering overload: %s", err)
	}
	
	// vm.RegisterCover("add")
	vm.RegisterBuiltin("p", p)
	vm.RegisterBuiltin("println", println)
	vm.RegisterBuiltin("printf", printf)
	vm.RegisterBuiltin("trace", trace)
	vm.RegisterBuiltin("val", val)
}


